(* 組み込み関数の類 *)

open Id
open Type

let funs_type_alist =
  let tv = new_tv () in [
  ["+"; "-"; "*"; "/"; "%"; "lsl"; "asr"], Tfun([Tint; Tint], Tint);
  ["+."; "-."; "*."; "/."], Tfun([Tfloat; Tfloat], Tfloat);
  ["sqrt"; "floor"; "fabs"; "sin"; "cos"; "atan"], Tfun([Tfloat], Tfloat);
  ["="; "<>"; "=="; "<"; "<="; ">"; ">="], (* != はパーサの都合で略 *)
  ForAll([tv], Tfun([Tvar tv; Tvar tv], Tbool));
  ["xor"; "xnor"; "&"; "or"], Tfun([Tbool; Tbool], Tbool); (* call by value *)
  ["not"], Tfun([Tbool], Tbool);
  ["print_char"], Tfun([Tint], Tunit);
  ["read_char"], Tfun([Tunit], Tint);
  ["create_array"], ForAll([tv], Tfun([Tint; Tvar tv], Tarray(Tvar tv)));
  ["<-"], ForAll([tv], Tfun([Tarray(Tvar tv); Tint; Tvar tv], Tunit));
  ["->"], (* 配列の参照 *)
  ForAll([tv], Tfun([Tarray(Tvar tv); Tint], Tvar tv));
  ["bits_of_float"; "int_of_float"], Tfun([Tfloat], Tint);
  ["float_of_bits"; "float_of_int"], Tfun([Tint], Tfloat);
]

let fun_idents =
  List.flatten
    (List.map
       (fun (xs,t) -> List.map (fun x -> new_id_t x t) xs)
       funs_type_alist)

(* 外部関数と名前が衝突すると内部的に面倒なのでちょっと変える *)
let fun_names =
  List.flatten
    (List.map
       (fun (xs,_) -> List.map
           (fun x -> if 'a' <= x.[0] && x.[0] <= 'z' then "." ^ x else x)
           xs)
       funs_type_alist) @
  ["t->"; "t<-"; ".tuple"; ".call_closure"]

let side_effective_funs = (* 副作用があるかもしれない関数 *)
  [".print_char"; ".read_char"; "<-"; "t<-"; ".call_closure"]
