(* クロージャ変換 *)
(* 1. 直接呼出し、クロージャなし
   2. 直接呼出し、クロージャあり
   3. クロージャ呼出し *)

open Global
open ILang

let result = ref S2.empty (* クロージャが作られる関数の集合 *)

(* env : 関数 -> クロージャがいるかどうか *)
let rec g env e =
  match e with
  | LetRec(x,ys,e1,e2) ->
      let l = Id.to_string x in
      let funs_bak = !gl_funs in
      (* クロージャがいらないと仮定して変換してみる *)
      let env_ = M.add x false env in
      let e1_ = g env_ e1 in
      let e2_ = g env_ e2 in
      let fv1 = S.remove_list ys (fv e1_) in
      let fv2 = fv e2_ in
      if S.is_empty fv1 && not (S.mem x fv2) then begin
        gl_funs := {
          name = l;
          args = ys;
          body = e1_;
        } :: !gl_funs;
        e2_
      end else begin
        (* 仮定が間違っていたのでやり直す *)
        Format.eprintf "closure for %s created@." l;
        result := S2.add l !result;
        gl_funs := funs_bak;
        let env_ = M.add x true env in
        let e1_ = g env_ e1 in
        let e2_ = g env_ e2 in
        let fv1 = S.elements (S.remove x (S.remove_list ys (fv e1_))) in
        if fv1 <> [] then
          Format.eprintf "free variable(s) %s found in function %s@."
            (Id.list_to_string fv1) l;
        let rec iter i fvs =
          match fvs with
          | [] -> e1_
          | z::zs -> Let(z, Op(Label "t->", [Var x; Int i]), iter (i+1) zs)
        in
        gl_funs := {
          name = l;
          args = x::ys;
          body = iter 1 fv1;
        } :: !gl_funs;
        Let(x, Op(Label ".tuple",
                  Label l :: List.map (fun x -> Var x) fv1),
            e2_)
      end
  | Op(Var x, ys) -> (* 必要なら引数リストの先頭をクロージャにする *)
      begin try
        if M.find x env then Op(Label (Id.to_string x), Var x :: ys)
        else Op(Label (Id.to_string x), ys)
      with Not_found -> Op(Label ".call_closure", Var x :: ys) end
  | If(c,x,y,e1,e2) -> If(c,x,y, g env e1, g env e2)
  | Let(x,e1,e2) -> Let(x, g env e1, g env e2)
  | Op(x,ys) -> Op(x,ys)
  | V x -> V x

let f e = g M.empty e
