(* 定数インデックスでしかアクセスされない配列を
   レジスタに割り付けちゃいますよ *)

open TreeAsm
open Global
open ILang

let result = Hashtbl.create 100

let remove env = function Label x -> M2.remove x env | _ -> env
let incr x env = try M2.add x (M2.find x env + 1) env with Not_found -> env

let rec check env e =
  match e with
  | V x -> remove env x
  | Op(Label "<-", [Label x; Int _; _])
  | Op(Label "->", [Label x; Int _]) -> incr x env
  | Op(_, xs) -> List.fold_left remove env xs
  | If(_,_,_,e1,e2) | Let(_,e1,e2) | LetRec(_,_,e1,e2) ->
      check (check env e1) e2

let sort env =
  let lst = M2.fold
      (fun x i lst -> (* 静的な出現回数で評価値をつける *)
        (x, float i /. float (Array.length (Hashtbl.find gl_vars x)))
        ::lst)
      env []
  in
  let lst = List.fast_sort (fun (_,x) (_,y) -> compare x y) lst in
  List.rev_map fst lst

let f e =
  Hashtbl.clear result;
  let m = Hashtbl.fold (fun k _ m -> M2.add k 0 m) gl_vars M2.empty in
  let m = Hashtbl.fold (fun _ v m -> Array.fold_left remove m v) gl_vars m in
  let m = List.fold_left (fun m { body = e } -> check m e) m !gl_funs in
  let m = check m e in
  let xs = sort m in
  let limit = Array.length regs_gl in
  let oft = ref 0 in
  List.iter
    (fun x ->
      let len = Array.length (Hashtbl.find gl_vars x) in
      if len + !oft <= limit then begin
        Hashtbl.add result x !oft;
        oft := !oft + len;
        Format.eprintf "%s allocated to register@." x end)
    xs;
  Format.eprintf "%d registers used for global variables@." !oft;
  e
