(* 部分冗長性除去 *)
(* 定数のロードを除去するためだけにあるといっても過言ではない *)

open Shared
open TreeAsm

module Exp = struct
  type t = exp
  let compare = compare
end

module Se = MySet.Make(Exp)
module Me = MyMap.Make(Exp)

let is_clean exp =
  match exp with
  | Op1(Wr,_) | Op1i(Ld,_,_) | Op2(Array,_,_) -> false
  | Imm _ | Op1 _ | Op1i _ | Op2 _ -> true
  | _ -> false

(* fst(返り値) はCallがあったかどうか *)
let rec used e =
  match e with
  | Ret exp -> used1 exp
  | Let(x,exp,e) -> let b,s = used1 exp in
    if b then (b,s)
    else let b2,s2 = used e in
    (b2, Se.union s s2)
and used1 exp =
  match exp with
  | If(_,_,_,e1,e2) | IfI(_,_,_,e1,e2) | FIf(_,_,_,e1,e2) ->
      let b1,s1 = used e1 and b2,s2 = used e2 in
      (b1 || b2, Se.inter s1 s2)
  | Call _ -> (true, Se.empty)
  | _ -> (false, if is_clean exp then Se.singleton exp else Se.empty)
let used e = snd (used e)

let add x s = if is_reg x then s else S2.add x s
let of_list xs = List.fold_left (fun s x -> add x s) S2.empty xs

let defable vs exp =
  let p x = x = reg_zero || x = reg_maxint || S2.mem x vs in
  match exp with
  | Imm _ -> true
  | Op1(_,x) | Op1i(_,x,_) -> p x
  | Op2(_,x,y) -> p x && p y
  | _ -> assert false

let rec g vs env e =
  match e with
  | Ret exp -> Ret(g1 vs env exp)
  | Let(x,exp,e_) ->
      let exxs = Se.fold
          (fun ex ls -> if Me.mem ex env then ls else (ex, new_str ()) :: ls)
          (Se.filter (defable vs) (used e)) []
      in
      let env1 = Me.add_list exxs env in
      List.fold_left
        (fun e (ex,x) -> Let(x,ex,e))
        (Let(x, g1 vs env1 exp, g (add x vs) env1 e_))
        exxs
and g1 vs env exp =
  match exp with
  | If(c,x,y,e1,e2) -> If(c,x,y, g vs env e1, g vs env e2)
  | IfI(c,x,y,e1,e2) -> IfI(c,x,y, g vs env e1, g vs env e2)
  | FIf(c,x,y,e1,e2) -> FIf(c,x,y, g vs env e1, g vs env e2)
  | _ -> try Mov(Me.find exp env) with Not_found -> exp

let rec f xs = List.map
    (fun x -> { x with body = g (of_list x.args) Me.empty x.body })
    xs
